#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script to train small networks on top of CXR embeddings.

`data_dir` is a path (absolute or relative) to a directory (can be a
Google Cloud bucket) containing TFRecord output from run_inference.py,
the generated embeddings for input CXRs. `labels_csv` is a CSV file with
format like:

image_id,split,AIRSPACE_OPACITY
004.png,train,0.0
001.png,train,0.0
000.png,tune,1.0

`image_id` corresponds to the image/id feature key within the TFRecord.
`split` is the name of the image set, e.g. 'training' or 'validation.'
This `split` is passed to one of `--train_split_name` or
`--tune_split_name`.

Example command:

`python3 -m train --train_split_name train --tune_split_name tune \
 --labels_csv labels.csv --head_name AIRSPACE_OPACITY \
 --data_dir ./data/ --num_epochs 30`
"""
from cxr_foundation import train_lib

import glob
import os
from absl import app
from absl import flags

import pandas as pd
import tensorflow as tf

flags.DEFINE_string('labels_csv', '', 'CSV file containing splits and labels')
flags.DEFINE_string(
    'data_dir', '', 'Absolute or relative path to directory '
    '(can be Google Cloud bucket starting with gs://) '
    'containing training data.')
flags.DEFINE_string(
    'train_split_name', '', 'Name of training image set '
    '(column `split`) in `labels_csv`.')
flags.DEFINE_string(
    'tune_split_name', '', 'Name of tune image set (column `split`) '
    'in `labels_csv`.')
flags.DEFINE_string(
    'head_name', None, 'Name of the head to train, '
    'e.g. the label class like AIRSPACE_OPACITY in the example above.')
flags.DEFINE_integer('batch_size', 512, 'Batch size.')
flags.DEFINE_integer('num_epochs', 300, 'Number of epochs to train.')

FLAGS = flags.FLAGS


def main(argv):
  with open(FLAGS.labels_csv) as f:
    df = pd.read_csv(f)
  df = df[~df[FLAGS.head_name].isna()]
  model = train_lib.create_model([FLAGS.head_name])
  training_df = df[df['split'] == FLAGS.train_split_name]
  training_labels = dict(
      zip(training_df['image_id'], training_df[FLAGS.head_name].astype(int)))
  filenames = glob.glob(os.path.join(FLAGS.data_dir, '*.tfrecord'))
  training_data = train_lib.get_dataset(filenames, labels=training_labels)
  tune_data = None
  if FLAGS.tune_split_name:
    tune_df = df[df['split'] == FLAGS.tune_split_name]
    tune_labels = dict(
        zip(tune_df['image_id'], tune_df[FLAGS.head_name].astype(int)))
    tune_data = train_lib.get_dataset(
        filenames, labels=tune_labels).batch(1).cache()
  model.fit(
      x=training_data.batch(FLAGS.batch_size).prefetch(
          tf.data.AUTOTUNE).cache(),
      validation_data=tune_data,
      epochs=FLAGS.num_epochs)
  model.save(
      os.path.join(FLAGS.data_dir, 'model'), include_optimizer=False)


if __name__ == '__main__':
  app.run(main)



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods to create tf.examples for model inference via pydicom."""

from cxr_foundation import constants

import io
from typing import Iterable, Union

import numpy as np
import png
import pydicom
from pydicom.pixel_data_handlers.util import apply_modality_lut
import tensorflow.compat.v1 as tf

_BITS_PER_BYTE = 8
_WINDOWWIDTH = 'WindowWidth'
_WINDOWCENTER = 'WindowCenter'


def _encode_png(array: np.ndarray) -> bytes:
  """Converts an unsigned integer 2-D NumPy array to a PNG-encoded string.

  Unsigned 8-bit and 16-bit images are supported.

  Args:
    array: Array to be encoded.

  Returns:
    PNG-encoded string.

  Raises:
    ValueError: If any of the following occurs:
      - `array` is not 2-D.
      - `array` data type is unsupported.
  """
  supported_types = frozenset([np.uint8, np.uint16])
  # Sanity checks.
  if array.ndim != 2:
    raise ValueError(f'Array must be 2-D. Actual dimensions: {array.ndim}')
  if array.dtype.type not in supported_types:
    raise ValueError('Pixels must be either `uint8` or `uint16`. '
                     f'Actual type: {array.dtype.name!r}')

  # Actual conversion.
  writer = png.Writer(
      width=array.shape[1],
      height=array.shape[0],
      greyscale=True,
      bitdepth=_BITS_PER_BYTE * array.dtype.itemsize)
  output_data = io.BytesIO()
  writer.write(output_data, array.tolist())
  return output_data.getvalue()


def _rescale_dynamic_range(image: np.ndarray) -> np.ndarray:
  """Rescales the dynamic range in an integer image to use the full bit range.

  Args:
    image: An image containing unsigned integer pixels.

  Returns:
    Rescaled copy of `image` that uses all the available bits per pixel.

  Raises:
    ValueError: If pixels are not of an integer type.
  """
  if not np.issubdtype(image.dtype, np.integer):
    raise ValueError('Image pixels must be an integer type. '
                     f'Actual type: {image.dtype.name!r}')
  iinfo = np.iinfo(image.dtype)
  return np.interp(image, (image.min(), image.max()),
                   (iinfo.min, iinfo.max)).astype(iinfo)


def _shift_to_unsigned(image: np.ndarray) -> np.ndarray:
  """Shifts values by the minimum value to an unsigned array suitible for PNG.

  This works with signed images and converts them to unsigned versions. It
  involves an inefficient step to convert to a larger data structure for
  shifting all values by the minimum value in the array. It also support float
  data by converting them into uint16.

  Args:
    image: An image containing signed integer pixels.

  Returns:
    Copy of `image` in an unsigned format. Note that the exact same image is
      returned when given an unsigned version.

  Raises:
    ValueError: If pixels are not of an integer type or float.
  """
  if image.dtype == np.uint16 or image.dtype == np.uint8:
    return image
  elif image.dtype == np.int16:
    image = image.astype(np.int32)
    return (image - np.min(image)).astype(np.uint16)
  elif image.dtype == np.int8:
    image = image.astype(np.int16)
    return (image - np.min(image)).astype(np.uint8)
  elif image.dtype == np.float:
    uint16_max = np.iinfo(np.uint16).max
    image = image - np.min(image)
    if np.max(image) > uint16_max:
      image = image * (uint16_max / np.max(image))
      image[image > uint16_max] = uint16_max
    return image.astype(np.uint16)
  raise ValueError('Image pixels must be an 8, 16 bit integer or float type. '
                   f'Actual type: {image.dtype.name!r}')


def _apply_pydicom_prep(ds: pydicom.Dataset) -> np.ndarray:
  """Prepares pixel data after applying data handling from pydicom."""

  def window_u16(image: np.ndarray, window_center: int,
                 window_width: int) -> np.ndarray:
    max_window = np.iinfo(np.uint16).max
    top_clip = window_center - 1 + window_width / 2
    bottom_clip = window_center - window_width / 2
    return np.interp(
        image.clip(bottom_clip, top_clip), (bottom_clip, top_clip),
        (0, max_window))

  arr = ds.pixel_array
  pixel_array = apply_modality_lut(arr, ds)
  if _WINDOWWIDTH in ds and _WINDOWCENTER in ds:
    window_center = ds.WindowCenter
    window_width = ds.WindowWidth
    if isinstance(ds.WindowCenter, pydicom.multival.MultiValue):
      window_center = int(ds.WindowCenter[0])
    if isinstance(ds.WindowWidth, pydicom.multival.MultiValue):
      window_width = int(ds.WindowWidth[0])
    pixel_array = window_u16(pixel_array, window_center, window_width)
  if ds.PhotometricInterpretation == 'MONOCHROME1':
    pixel_array = np.max(pixel_array) - pixel_array
  pixel_array = _shift_to_unsigned(pixel_array)
  # Don't rescale dynamic range for 8-bit images like CXR14.
  if pixel_array.dtype != np.uint8:
    pixel_array = _rescale_dynamic_range(pixel_array)
  return pixel_array


def _assign_bytes_feature(feature: tf.train.Feature,
                          value: Union[bytes, Iterable[bytes]]) -> None:
  """Assigns a bytes float value into feature."""
  if isinstance(value, bytes):
    feature.bytes_list.value[:] = [value]
  else:
    assert not isinstance(value, str)
    feature.bytes_list.value[:] = list(value)


def png_to_tfexample(image_array: np.ndarray) -> tf.train.Example:
  """Create a tf.example for inference.

  The image will be spread to the full bit-depth of 16-bit images.

  Args:
    image_array: An image to use to create the example.

  Returns:
    example: A tf.example for inference.
  """
  pixel_array = _shift_to_unsigned(image_array)
  # Don't rescale dynamic range for 8-bit images like CXR14.
  if pixel_array.dtype != np.uint8:
    pixel_array = _rescale_dynamic_range(pixel_array)
  png_bytes = _encode_png(pixel_array)
  example = tf.train.Example()
  features = example.features.feature
  _assign_bytes_feature(features[constants.IMAGE_KEY], png_bytes)
  _assign_bytes_feature(features[constants.IMAGE_FORMAT_KEY], b'png')
  return example


def dicom_to_tfexample(single_dicom: pydicom.Dataset) -> tf.train.Example:
  """Create a tf.example for inference.

  Resulting images are spread to the full bit-depth of 16-bit images.
  Applies apply_modality_lut first followed by window/level if prresent.

  Args:
    single_dicom: A pydicom dataset used to create the example.

  Returns:
    example: A tf.example for inference.
  """
  image_array = _apply_pydicom_prep(single_dicom)
  png_bytes = _encode_png(image_array)
  example = tf.train.Example()
  features = example.features.feature
  _assign_bytes_feature(features[constants.IMAGE_KEY], png_bytes)
  _assign_bytes_feature(features[constants.IMAGE_FORMAT_KEY], b'png')
  return example



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Beam pipeline to generate embeddings from a list of PNG or DICOM files."""
from cxr_foundation import constants
from cxr_foundation import inference_beam_lib

import argparse
import glob
import logging
import os

import apache_beam as beam
from apache_beam.options import pipeline_options


def main(argv=None, save_main_session=True):
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--input_path',
      dest='input_path',
      required=True,
      help='Folder or GCS bucket containing input PNG or DICOM files.')
  parser.add_argument(
      '--input_file_type',
      dest='input_file_type',
      required=True,
      default=inference_beam_lib.InputFileType.PNG,
      type=inference_beam_lib.InputFileType,
      choices=list(inference_beam_lib.InputFileType),
      help=f'The type of input files, either DICOM or PNG.')
  parser.add_argument(
      '--output_path',
      dest='output_path',
      required=True,
      help='Folder or GCS bucket to write output tf.Examples.')
  parser.add_argument(
      '--limit',
      dest='limit',
      type=int,
      help='Only process this many input files.')
  parser.add_argument(
      '--endpoint_id',
      dest='endpoint_id',
      type=int,
      default=6695981832690728960,
      help='Numerical ID for embeddings API.')
  parser.add_argument(
      '--embeddings_project',
      dest='embeddings_project',
      default='gh-rad-validation-cxrembd-deid',
      help='GCP project ID that hosts embeddings API.')
  parser.add_argument(
      '--skip_errors',
      dest='skip_errors',
      default=False,
      help='Whether to suppress exceptions thrown by GenerateEmbeddingsDoFn.')
  known_args, pipeline_args = parser.parse_known_args(argv)
  options = pipeline_options.PipelineOptions(pipeline_args)

  if known_args.input_path.startswith(constants.GCS_PREFIX):
    input = list(beam.io.gcp.gcsio.GcsIO().list_prefix(known_args.input_path))
  elif '*' in known_args.input_path:
    input = glob.glob(known_args.input_path)
  else:
    input = [
        os.path.join(known_args.input_path, f)
        for f in os.listdir(known_args.input_path)
    ]

  if (not known_args.output_path.startswith(constants.GCS_PREFIX) and
      not os.path.exists(known_args.output_path)):
    os.mkdir(known_args.output_path)

  if known_args.limit > 0:
    input = input[:known_args.limit]

  options.view_as(
      pipeline_options.SetupOptions).save_main_session = save_main_session

  with beam.Pipeline(options=options) as p:
    _ = (
        p | beam.Create(input) | beam.ParDo(
            inference_beam_lib.CreateExampleDoFn(
                output_path=known_args.output_path,
                input_file_type=known_args.input_file_type,
                skip_errors=known_args.skip_errors))
        | beam.ParDo(
            inference_beam_lib.GenerateEmbeddingsDoFn(
                known_args.embeddings_project, known_args.endpoint_id,
                skip_errors=known_args.skip_errors))
        | beam.ParDo(
            inference_beam_lib.ProcessPredictionDoFn(
                output_path=known_args.output_path))
        | beam.Map(print))


if __name__ == '__main__':
  main()



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

GCS_PREFIX = 'gs://'
# Vertex is only in us-central1 for now.
LOCATION = 'us-central1'

# tf.Example feature keys.
IMAGE_KEY = 'image/encoded'
IMAGE_FORMAT_KEY = 'image/format'
IMAGE_ID_KEY = 'image/id'
EMBEDDING_KEY = 'embedding'



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import constants
import example_generator_lib
import unittest
from unittest import mock

import apache_beam as beam
from apache_beam.testing import test_pipeline
from apache_beam.testing import util as test_util
import functools
import numpy as np
import os
import six
import tempfile
import tensorflow.compat.v1 as tf

import inference_beam_lib


class InferenceBeamTestCase(unittest.TestCase):

  def assertMetricsEqual(self, metrics, metric_name, expected):
    actual = metrics.query(beam.metrics.MetricsFilter().with_name(
        metric_name))['counters'][0].committed
    self.assertEqual(expected, actual)


class CreateExampleTest(InferenceBeamTestCase):

  def setUp(self):
    self._output_path = tempfile.TemporaryDirectory()
    self._input_file = './testdata/random.png'
    self._expected = tf.train.Example()
    with open('./testdata/expected.png', 'rb') as f:
      self._expected.features.feature[constants.IMAGE_KEY].bytes_list.value[:] = [f.read()]
    self._expected.features.feature[constants.IMAGE_ID_KEY].bytes_list.value[:] = [
        six.ensure_binary(self._input_file)
    ]
    self._expected.features.feature[constants.IMAGE_FORMAT_KEY].bytes_list.value[:] = [
      b'png'
    ]

  def tearDown(self):
    self._output_path.cleanup()

  def test_success(self):
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([self._input_file])
      output = input | beam.ParDo(
          inference_beam_lib.CreateExampleDoFn(self._output_path.name))
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([self._expected]))

  def test_already_existing(self):
    filebase = inference_beam_lib._image_id_to_filebase(self._input_file)
    with open(
        os.path.join(self._output_path.name, f'{filebase}.tfrecord'), 'w') as f:
      f.write('Output file already exists.')
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([self._input_file])
      output = input | beam.ParDo(
          inference_beam_lib.CreateExampleDoFn(self._output_path.name))
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([]))
      self.assertMetricsEqual(result.metrics(), 'output-file-exists', 1)


class GenerateEmbeddingsTest(InferenceBeamTestCase):

  def setUp(self):
    self._input_example = tf.train.Example()
    self._input_example.features.feature[
        constants.IMAGE_KEY].bytes_list.value[:] = [b'PNGblahblahblah']

  @mock.patch('google.cloud.aiplatform.gapic.PredictionServiceClient')
  def test_success(self, mock_api_client):
    mock_api_client.return_value.predict.return_value.predictions = [
        1.0, 0.0, 3.0
    ]
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([self._input_example])
      output = input | beam.ParDo(
          inference_beam_lib.GenerateEmbeddingsDoFn(
              'project',
              'endpoint')) | beam.FlatMap(lambda x: x.features.feature[
                  constants.EMBEDDING_KEY].float_list.value)
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([1.0, 0.0, 3.0]))
      self.assertMetricsEqual(result.metrics(), 'successful-inference', 1)

  def test_missing_image_feature(self):
    input_example = tf.train.Example()
    input_example.features.feature['unused'].bytes_list.value[:] = [b'unused']
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([input_example])
      output = input | beam.ParDo(
          inference_beam_lib.GenerateEmbeddingsDoFn('project', 'endpoint'))
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([]))
      self.assertMetricsEqual(result.metrics(), 'skipped-inference', 1)

  @mock.patch('google.cloud.aiplatform.gapic.PredictionServiceClient')
  def test_raise_error(self, mock_api_client):

    def failed_inference(*args, **kwargs):
      raise ValueError

    mock_api_client.return_value.predict.side_effect = failed_inference
    with self.assertRaises(RuntimeError):
      with test_pipeline.TestPipeline() as p:
        input = p | beam.Create([self._input_example])
        output = input | beam.ParDo(
            inference_beam_lib.GenerateEmbeddingsDoFn(
                'project', 'endpoint', skip_errors=False))
        p.run()

  @mock.patch('google.cloud.aiplatform.gapic.PredictionServiceClient')
  def test_skip_error(self, mock_api_client):

    def failed_inference(*args, **kwargs):
      raise ValueError

    mock_api_client.return_value.predict.side_effect = failed_inference
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([self._input_example])
      output = input | beam.ParDo(
          inference_beam_lib.GenerateEmbeddingsDoFn(
              'project', 'endpoint', skip_errors=True))
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([None]))
      self.assertMetricsEqual(result.metrics(), 'failed-inference', 1)


class ProcessPredictionTest(InferenceBeamTestCase):

  def setUp(self):
    self._output_path = tempfile.TemporaryDirectory()

  def tearDown(self):
    self._output_path.cleanup()

  def test_missing_element(self):
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([None])
      output = input | beam.ParDo(inference_beam_lib.ProcessPredictionDoFn())
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([]))
      self.assertMetricsEqual(result.metrics(), 'missing-representation', 1)

  def test_missing_representation(self):
    input_example = tf.train.Example()
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([input_example])
      output = input | beam.ParDo(inference_beam_lib.ProcessPredictionDoFn())
      result = p.run()
      test_util.assert_that(output, test_util.equal_to([]))
      self.assertMetricsEqual(result.metrics(), 'missing-representation', 1)

  def test_no_output_path(self):
    input_example = tf.train.Example()
    input_example.features.feature[
        constants.EMBEDDING_KEY].float_list.value[:] = [1.0, 2.0, 3.0]
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([input_example])
      output = input | beam.ParDo(inference_beam_lib.ProcessPredictionDoFn())
      result = p.run()
      test_util.assert_that(
          output,
          functools.partial(
              np.testing.assert_array_almost_equal, y=[[1.0, 2.0, 3.0]]))

  def test_output_path(self):
    input_example = tf.train.Example()
    input_example.features.feature[
        constants.IMAGE_ID_KEY].bytes_list.value[:] = [b'image1234.png']
    input_example.features.feature[
        constants.EMBEDDING_KEY].float_list.value[:] = [1.0, 2.0, 3.0]
    with test_pipeline.TestPipeline() as p:
      input = p | beam.Create([input_example])
      output = input | beam.ParDo(
          inference_beam_lib.ProcessPredictionDoFn(self._output_path.name))
      result = p.run()
      test_util.assert_that(
          output,
          functools.partial(
              np.testing.assert_array_almost_equal, y=[[1.0, 2.0, 3.0]]))
      actual = tf.train.Example()
      filename = os.path.join(self._output_path.name, 'image1234.tfrecord')
      for example in tf.python_io.tf_record_iterator(filename):
        actual = tf.train.Example.FromString(example)
      self.assertEqual(input_example, actual)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import train_lib

import unittest

import glob
import numpy as np
import tensorflow.compat.v1 as tf

from typing import Any, List, Tuple


def _flatten(dataset: List[Tuple[tf.Tensor]]) -> List[Any]:
  flattened = []
  for datum in dataset:
    row = []
    for i in datum:
      try:
        row.append(list(i))
      except TypeError:
        row.append(i)
    flattened.append(row)
  return flattened


class TrainLibTestCase(unittest.TestCase):

  def test_empty_labels_fails(self):
    with self.assertRaises(AssertionError):
      train_lib.get_dataset(sorted(glob.glob('./testdata/*.tfrecord')), {})

  def test_get_dataset_single_label(self):
    dataset = _flatten(
        train_lib.get_dataset(
            sorted(glob.glob('./testdata/*.tfrecord')),
            {'gs://superrad/inputs/cxr14/00000001_000.png': 1},
            embeddings_size=5))
    actual_embeddings = np.array([d[0] for d in dataset])
    actual_labels = np.array([d[1] for d in dataset])
    actual_weights = np.array([d[2] for d in dataset])
    np.testing.assert_array_almost_equal(
        np.array([[3., 0., 4., 0., 3.]]), actual_embeddings)
    np.testing.assert_array_almost_equal(
        np.array([[1]]), actual_labels)
    np.testing.assert_array_almost_equal(np.ones(1), actual_weights)

  def test_get_dataset_weights(self):
    dataset = _flatten(
        train_lib.get_dataset(
            sorted(glob.glob('./testdata/*.tfrecord')),
            {'gs://superrad/inputs/cxr14/00000001_000.png': 1},
            embeddings_size=5,
            weights={'gs://superrad/inputs/cxr14/00000001_000.png': 0.5}))
    actual_embeddings = np.array([d[0] for d in dataset])
    actual_labels = np.array([d[1] for d in dataset])
    actual_weights = np.array([d[2] for d in dataset])
    np.testing.assert_array_almost_equal(
        np.array([[3., 0., 4., 0., 3.]]), actual_embeddings)
    np.testing.assert_array_almost_equal(np.array([[1]]), actual_labels)
    np.testing.assert_array_almost_equal(np.array([[0.5]]), actual_weights)



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Collection of Beam DoFns to generate embeddings."""
from cxr_foundation import constants
from cxr_foundation import example_generator_lib

import base64
import contextlib
from enum import Enum
import io
from apache_beam.utils import retry

import os
import six
import tempfile
import numpy as np
from PIL import Image
import pydicom

import apache_beam as beam

from apache_beam.options import pipeline_options
from apache_beam.options.pipeline_options import GoogleCloudOptions
from apache_beam.runners import DataflowRunner
import google.auth

import tensorflow.compat.v1 as tf

from google.api_core import exceptions
from google.api_core.client_options import ClientOptions
from google.api_core.retry import Retry
from google.cloud import aiplatform

from typing import Any, List, Mapping, Optional, Sequence, Text

_RETRIABLE_TYPES = (
    exceptions.TooManyRequests,  # HTTP 429
    exceptions.InternalServerError,  # HTTP 500
    exceptions.BadGateway,  # HTTP 502
    exceptions.ServiceUnavailable,  # HTTP 503
    exceptions.DeadlineExceeded,  # HTTP 504
)

_METRICS_NAMESPACE = 'cxr-embeddings'
_API_ENDPOINT = 'us-central1-aiplatform.googleapis.com'
_VIEW_POSITION = 'ViewPosition'
_FRONTAL_VIEW_POSITIONS = ('AP', 'PA')


class InputFileType(Enum):
  PNG = 'png'
  DICOM = 'dicom'

  def __str__(self):
      return self.value


def _is_retryable(exc):
  return isinstance(exc, _RETRIABLE_TYPES)


def _image_id_to_filebase(image_id: str) -> str:
  filebase, _ = os.path.splitext(os.path.basename(image_id))
  return filebase


@contextlib.contextmanager
def _open(path, *args, **kwargs):
  if path.startswith(constants.GCS_PREFIX):
    f = beam.io.gcp.gcsio.GcsIO().open(path, *args, **kwargs)
  else:
    f = open(path, *args, **kwargs)
  yield f
  f.close()


@beam.typehints.with_input_types(str)
@beam.typehints.with_output_types(Optional[tf.train.Example])
class CreateExampleDoFn(beam.DoFn):

  def __init__(self,
               output_path: Optional[str] = None,
               input_file_type: InputFileType = InputFileType.PNG,
               skip_errors: bool = False):
    self._output_path = output_path
    self._input_file_type = input_file_type
    self._skip_errors = skip_errors

  def process(self, uri: str):
    if self._output_path:
      exists = False
      filebase = _image_id_to_filebase(uri)
      filename = f'{self._output_path}/{filebase}.tfrecord'
      if self._output_path.startswith(constants.GCS_PREFIX):
        exists = beam.io.gcp.gcsio.GcsIO().exists(filename)
      else:
        exists = os.path.exists(filename)
      if exists:
        beam.metrics.Metrics.counter(_METRICS_NAMESPACE,
                                     'output-file-exists').inc()
        return
    try:
      with _open(uri, 'rb') as f:
        if self._input_file_type == InputFileType.PNG:
          img = np.asarray(Image.open(io.BytesIO(f.read())).convert('L'))
          example = example_generator_lib.png_to_tfexample(img)
        elif self._input_file_type == InputFileType.DICOM:
          dicom = pydicom.dcmread(io.BytesIO(f.read()))
          if _VIEW_POSITION in dicom and dicom.ViewPosition not in _FRONTAL_VIEW_POSITIONS:
            beam.metrics.Metrics.counter(_METRICS_NAMESPACE, 'non-frontal-cxr').inc()
            return
          example = example_generator_lib.dicom_to_tfexample(dicom)
        else:
          raise ValueError('Unknown file type.')
        example.features.feature[constants.IMAGE_ID_KEY].bytes_list.value[:] = [
            six.ensure_binary(uri)
        ]
      yield example
    except Exception as e:
      beam.metrics.Metrics.counter(_METRICS_NAMESPACE, 'failed-example-creation').inc()
      if not self._skip_errors:
        raise e


@beam.typehints.with_input_types(Optional[tf.train.Example])
@beam.typehints.with_output_types(Optional[tf.train.Example])
class GenerateEmbeddingsDoFn(beam.DoFn):
  """A DoFn that generates embeddings from a cloud-hosted TensorFlow model."""

  def __init__(self, project, endpoint_id, skip_errors: bool = False):
    self._project = project
    self._endpoint_id = endpoint_id
    self._skip_errors = skip_errors

  def _make_instances(self,
                      serialized_example: bytes) -> List[Mapping[Text, Any]]:
    return [{'b64': base64.b64encode(serialized_example).decode()}]

  def _run_inference(self, serialized_example: bytes) -> Sequence[float]:
    instances = self._make_instances(serialized_example)
    api_client = aiplatform.gapic.PredictionServiceClient(
        client_options=ClientOptions(api_endpoint=_API_ENDPOINT))
    endpoint = api_client.endpoint_path(
        project=self._project,
        location=constants.LOCATION,
        endpoint=self._endpoint_id)
    retry_policy = Retry(predicate=_is_retryable)
    try:
      response = api_client.predict(
          endpoint=endpoint, instances=instances, retry=retry_policy, timeout=60)
      beam.metrics.Metrics.counter(_METRICS_NAMESPACE,
                                   'successful-inference').inc()
    except Exception as e:
      beam.metrics.Metrics.counter(_METRICS_NAMESPACE, 'failed-inference').inc()
      if not self._skip_errors:
        raise e
      return []
    return response.predictions

  def process(self, example: Optional[tf.train.Example]):
    if not example or constants.IMAGE_KEY not in example.features.feature:
      beam.metrics.Metrics.counter(_METRICS_NAMESPACE,
                                   'skipped-inference').inc()
      return
    serialized_example = example.SerializeToString()
    outputs = self._run_inference(serialized_example)
    yield self._post_process(serialized_example, outputs)

  def _post_process(self, serialized_example: bytes,
                    outputs: Sequence[float]) -> Optional[tf.train.Example]:
    example = tf.train.Example()
    example.ParseFromString(serialized_example)
    if not outputs:
      return
    values = np.array(outputs)
    example.features.feature[
        constants.EMBEDDING_KEY].float_list.value[:] = values.flatten()
    return example


@beam.typehints.with_input_types(Optional[tf.train.Example])
@beam.typehints.with_output_types(np.ndarray)
class ProcessPredictionDoFn(beam.DoFn):

  def __init__(self, output_path: Optional[str] = None):
    self._output_path = output_path

  def process(self, element: Optional[tf.train.Example]):
    if not element or constants.EMBEDDING_KEY not in element.features.feature:
      beam.metrics.Metrics.counter(_METRICS_NAMESPACE,
                                   'missing-representation').inc()
      return np.array([])
    if constants.IMAGE_KEY in element.features.feature:
      del element.features.feature[constants.IMAGE_KEY]
    if self._output_path:
      filebase = _image_id_to_filebase(
          six.ensure_str(element.features.feature[
              constants.IMAGE_ID_KEY].bytes_list.value[0]))
      filename = f'{self._output_path}/{filebase}.tfrecord'
      with tempfile.NamedTemporaryFile('wb+', delete=False) as f:
        with tf.io.TFRecordWriter(f.name) as w:
          w.write(element.SerializeToString())
        with _open(filename, 'wb') as o:
          o.write(f.read())
    yield np.array(
        element.features.feature[constants.EMBEDDING_KEY].float_list.value[:])



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library functions for training small networks on CXR embeddings."""
from cxr_foundation import constants

import functools
import glob

import tensorflow.compat.v2 as tf
import tensorflow_models as tfm

from typing import Dict, List, Optional, Tuple


_DEFAULT_EMBEDDINGS_SIZE = 1376


def parse_fn(
    serialized_example: bytes,
    embeddings_size: int = _DEFAULT_EMBEDDINGS_SIZE,
    example_feature_key: str = constants.IMAGE_ID_KEY,
    embeddings_feature_key: str = constants.EMBEDDING_KEY
) -> Tuple[tf.Tensor, tf.Tensor]:
  """Parses a single tf.Example to return embeddings + label/weights."""
  features = {
      embeddings_feature_key:
          tf.io.FixedLenFeature([embeddings_size],
                                tf.float32,
                                default_value=tf.constant(
                                    0.0, shape=[embeddings_size]))
  }
  features[example_feature_key] = tf.io.FixedLenSequenceFeature(
      [], tf.string, allow_missing=True, default_value='')
  parsed_tensors = tf.io.parse_example(serialized_example, features=features)
  return (parsed_tensors[embeddings_feature_key],
          parsed_tensors[example_feature_key])


def process_tfrecord_shard(
    filename: str,
    embeddings_size: int = _DEFAULT_EMBEDDINGS_SIZE,
    example_feature_key: str = constants.IMAGE_ID_KEY,
    embeddings_feature_key: str = constants.EMBEDDING_KEY) -> tf.data.Dataset:
  """Process a single shard of a TFRecord."""
  return tf.data.TFRecordDataset(filename).map(
      functools.partial(
          parse_fn,
          embeddings_size=embeddings_size,
          example_feature_key=example_feature_key,
          embeddings_feature_key=embeddings_feature_key))


def get_dataset(
    filenames: List[str],
    labels: Dict[str, int],
    embeddings_size: int = _DEFAULT_EMBEDDINGS_SIZE,
    weights: Optional[Dict[str, float]] = None,
    example_feature_key: str = constants.IMAGE_ID_KEY,
    embeddings_feature_key: str = constants.EMBEDDING_KEY) -> tf.data.Dataset:
  """Create tf.data.Dataset from the TFRecords."""
  assert labels, 'Must pass non-empty labels dict.'
  dataset = tf.data.Dataset.from_tensor_slices(filenames).interleave(
      functools.partial(
          process_tfrecord_shard,
          embeddings_size=embeddings_size,
          example_feature_key=example_feature_key,
          embeddings_feature_key=embeddings_feature_key))

  labels = tf.lookup.StaticHashTable(
      tf.lookup.KeyValueTensorInitializer(
          tf.constant(list(labels.keys())),
          tf.constant(list(labels.values())),
          key_dtype=tf.string,
          value_dtype=tf.int32),
      default_value=-1)

  if weights is not None:
    weights = tf.lookup.StaticHashTable(
        tf.lookup.KeyValueTensorInitializer(
            tf.constant(list(weights.keys())),
            tf.constant(list(weights.values())),
            key_dtype=tf.string,
            value_dtype=tf.float32),
        default_value=0.)

  def filter_zero_weights(embedding, label, weight):
    del embedding, label
    return tf.math.reduce_all(weight > 0)

  def filter_negative_labels(embedding, label, weight):
    return tf.math.reduce_all(label > -1)

  def _lookup_label(features, image_id):
    return features, labels.lookup(image_id), image_id

  def _lookup_weight(features, label, image_id):
    weight = 1.0
    if weights:
      weight = weights.lookup(image_id)
    return features, label, weight, image_id

  # Look up labels and weights.
  dataset = dataset.map(_lookup_label).map(_lookup_weight)

  # Remove the IDs and filter out zero weights and negative labels.
  return dataset.map(lambda features, label, weight, image_id:
                     (features, label, weight)).filter(
                         filter_negative_labels).filter(filter_zero_weights)


def create_model(heads,
                 embeddings_size = _DEFAULT_EMBEDDINGS_SIZE,
                 learning_rate=0.1,
                 end_lr_factor=1.0,
                 dropout=0.0,
                 decay_steps=1000,
                 loss_weights=None,
                 hidden_layer_sizes=[512, 256],
                 weight_decay=0.0,
                 seed=None):
  """Creates linear probe or multilayer perceptron using LARS + cosine decay."""
  inputs = tf.keras.Input(shape=(embeddings_size,))
  hidden = inputs
  # If no hidden_layer_sizes are provided, model will be a linear probe.
  for size in hidden_layer_sizes:
    hidden = tf.keras.layers.Dense(
        size,
        activation='relu',
        kernel_initializer=tf.keras.initializers.HeUniform(seed=seed),
        kernel_regularizer=tf.keras.regularizers.l2(l2=weight_decay),
        bias_regularizer=tf.keras.regularizers.l2(l2=weight_decay))(
            hidden)
    hidden = tf.keras.layers.BatchNormalization()(hidden)
    hidden = tf.keras.layers.Dropout(dropout, seed=seed)(hidden)
  output = tf.keras.layers.Dense(
      units=len(heads),
      activation='sigmoid',
      kernel_initializer=tf.keras.initializers.HeUniform(seed=seed))(
          hidden)
  outputs = {}
  for i, head in enumerate(heads):
    outputs[head] = tf.keras.layers.Lambda(
        lambda x: x[..., i:i + 1], name=head.lower())(
            output)
  model = tf.keras.Model(inputs, outputs)
  learning_rate_fn = tf.keras.experimental.CosineDecay(
      tf.cast(learning_rate, tf.float32),
      tf.cast(decay_steps, tf.float32),
      alpha=tf.cast(end_lr_factor, tf.float32))
  model.compile(
      optimizer=tfm.optimization.lars_optimizer.LARS(
          learning_rate=learning_rate_fn),
      loss=dict([(head, 'binary_crossentropy') for head in heads]),
      loss_weights=loss_weights or dict([(head, 1.) for head in heads]),
      weighted_metrics=['AUC'])
  return model



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools

setuptools.setup(
  name='cxr-foundation',
  version='0.0.12',
  description='CXR Foundation: chest x-ray embeddings generation.',
  install_requires=[
    'google-api-python-client',
    'google-apitools',
    'google-cloud-aiplatform',
    'google-cloud-storage',
    'apache_beam',
    'pandas',
    'tensorflow >= 2.10.0',
    'pillow',
    'pypng',
    'pydicom',
    'tf-models-official >= 2.10.0',
    'protobuf',
    'typing-extensions',
    'shapely < 2.0.0'
  ],
  packages=setuptools.find_packages())



#!/usr/bin/python
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for example_generator_lib."""
import example_generator_lib

import numpy as np
import pydicom
import unittest


class ExampleGeneratorLibTest(unittest.TestCase):

  def test_create_example(self):
    """Test the creation of examples."""
    # This is a DICOM with a grayscale fake image.
    dicom_path = './testdata/fake.dcm'
    dicom = pydicom.dcmread(dicom_path)

    test_example = example_generator_lib.dicom_to_tfexample(dicom)
    f_dict = test_example.features.feature
    self.assertEqual(f_dict['image/format'].bytes_list.value[:], [b'png'])
    self.assertEqual(len(f_dict['image/encoded'].bytes_list.value[0]), 23287)


if __name__ == '__main__':
  unittest.main()